from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class ShardPerformanceSummary(_message.Message):
    __slots__ = ("operation_id", "shard_id", "performance_summary_with_query_config")
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    SHARD_ID_FIELD_NUMBER: _ClassVar[int]
    PERFORMANCE_SUMMARY_WITH_QUERY_CONFIG_FIELD_NUMBER: _ClassVar[int]
    operation_id: str
    shard_id: int
    performance_summary_with_query_config: _struct_pb2.Struct
    def __init__(
        self,
        operation_id: _Optional[str] = ...,
        shard_id: _Optional[int] = ...,
        performance_summary_with_query_config: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...,
    ) -> None: ...

class ListOfflineQueryShardPerformanceSummariesPageToken(_message.Message):
    __slots__ = ("shard_id_hwm",)
    SHARD_ID_HWM_FIELD_NUMBER: _ClassVar[int]
    shard_id_hwm: int
    def __init__(self, shard_id_hwm: _Optional[int] = ...) -> None: ...

class ListOfflineQueryShardPerformanceSummariesRequest(_message.Message):
    __slots__ = ("operation_id", "limit", "page_token")
    OPERATION_ID_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    operation_id: str
    limit: int
    page_token: str
    def __init__(
        self, operation_id: _Optional[str] = ..., limit: _Optional[int] = ..., page_token: _Optional[str] = ...
    ) -> None: ...

class ListOfflineQueryShardPerformanceSummariesResponse(_message.Message):
    __slots__ = ("performance_summaries", "next_page_token")
    PERFORMANCE_SUMMARIES_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    performance_summaries: _containers.RepeatedCompositeFieldContainer[ShardPerformanceSummary]
    next_page_token: str
    def __init__(
        self,
        performance_summaries: _Optional[_Iterable[_Union[ShardPerformanceSummary, _Mapping]]] = ...,
        next_page_token: _Optional[str] = ...,
    ) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/common/v1/query_log.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.common.v1 import operation_kind_pb2 as chalk_dot_common_dot_v1_dot_operation__kind__pb2
from chalk._gen.chalk.common.v1 import query_status_pb2 as chalk_dot_common_dot_v1_dot_query__status__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1f\x63halk/common/v1/query_log.proto\x12\x0f\x63halk.common.v1\x1a$chalk/common/v1/operation_kind.proto\x1a"chalk/common/v1/query_status.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\x81\x01\n\x12VersionedQueryName\x12\x1d\n\nquery_name\x18\x01 \x01(\tR\tqueryName\x12\x31\n\x12query_name_version\x18\x02 \x01(\tH\x00R\x10queryNameVersion\x88\x01\x01:\x02\x18\x01\x42\x15\n\x13_query_name_version"\xf3\x03\n\x0fQueryLogFilters\x12!\n\x0coperation_id\x18\x01 \x03(\tR\x0boperationId\x12\x45\n\x0eoperation_kind\x18\x02 \x03(\x0e\x32\x1e.chalk.common.v1.OperationKindR\roperationKind\x12\x42\n\nquery_name\x18\x03 \x03(\x0b\x32#.chalk.common.v1.VersionedQueryNameR\tqueryName\x12\x19\n\x08\x61gent_id\x18\x05 \x03(\tR\x07\x61gentId\x12\x1f\n\x0b\x62ranch_name\x18\x06 \x03(\tR\nbranchName\x12%\n\x0e\x63orrelation_id\x18\x07 \x03(\tR\rcorrelationId\x12\x19\n\x08trace_id\x18\x08 \x03(\tR\x07traceId\x12"\n\rquery_plan_id\x18\t \x03(\tR\x0bqueryPlanId\x12#\n\rdeployment_id\x18\n \x03(\tR\x0c\x64\x65ploymentId\x12?\n\x0cquery_status\x18\x0b \x03(\x0e\x32\x1c.chalk.common.v1.QueryStatusR\x0bqueryStatus\x12&\n\x0fmeta_query_hash\x18\x0c \x03(\tR\rmetaQueryHash:\x02\x18\x01"\x97\x01\n\x1bGetQueryLogEntriesPageToken\x12(\n\x10operation_id_hwm\x18\x01 \x01(\tR\x0eoperationIdHwm\x12J\n\x13query_timestamp_hwm\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x11queryTimestampHwm:\x02\x18\x01"\xa2\x03\n\x19GetQueryLogEntriesRequest\x12l\n%query_timestamp_lower_bound_inclusive\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR!queryTimestampLowerBoundInclusive\x12q\n%query_timestamp_upper_bound_exclusive\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00R!queryTimestampUpperBoundExclusive\x88\x01\x01\x12:\n\x07\x66ilters\x18\x03 \x01(\x0b\x32 .chalk.common.v1.QueryLogFiltersR\x07\x66ilters\x12\x1b\n\tpage_size\x18\x04 \x01(\x05R\x08pageSize\x12\x1d\n\npage_token\x18\x05 \x01(\tR\tpageToken:\x02\x18\x01\x42(\n&_query_timestamp_upper_bound_exclusive"\xa7\x06\n\rQueryLogEntry\x12!\n\x0coperation_id\x18\x01 \x01(\tR\x0boperationId\x12%\n\x0e\x65nvironment_id\x18\x02 \x01(\tR\renvironmentId\x12#\n\rdeployment_id\x18\x03 \x01(\tR\x0c\x64\x65ploymentId\x12\x45\n\x0eoperation_kind\x18\x04 \x01(\x0e\x32\x1e.chalk.common.v1.OperationKindR\roperationKind\x12\x43\n\x0fquery_timestamp\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0equeryTimestamp\x12L\n\x14\x65xecution_started_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x12\x65xecutionStartedAt\x12N\n\x15\x65xecution_finished_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x13\x65xecutionFinishedAt\x12?\n\x0cquery_status\x18\x08 \x01(\x0e\x32\x1c.chalk.common.v1.QueryStatusR\x0bqueryStatus\x12\x1d\n\nquery_name\x18\t \x01(\tR\tqueryName\x12,\n\x12query_name_version\x18\n \x01(\tR\x10queryNameVersion\x12\x19\n\x08\x61gent_id\x18\x0b \x01(\tR\x07\x61gentId\x12\x1f\n\x0b\x62ranch_name\x18\x0c \x01(\tR\nbranchName\x12%\n\x0e\x63orrelation_id\x18\r \x01(\tR\rcorrelationId\x12\x19\n\x08trace_id\x18\x0e \x01(\tR\x07traceId\x12"\n\rquery_plan_id\x18\x0f \x01(\tR\x0bqueryPlanId\x12!\n\x0cvalue_tables\x18\x10 \x03(\tR\x0bvalueTables\x12&\n\x0fmeta_query_hash\x18\x11 \x01(\tR\rmetaQueryHash:\x02\x18\x01"\x82\x01\n\x1aGetQueryLogEntriesResponse\x12\x38\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x1e.chalk.common.v1.QueryLogEntryR\x07\x65ntries\x12&\n\x0fnext_page_token\x18\x02 \x01(\tR\rnextPageToken:\x02\x18\x01\x42\x82\x01\n\x13\x63om.chalk.common.v1B\rQueryLogProtoP\x01\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43halk.Common.V1\xca\x02\x0f\x43halk\\Common\\V1\xe2\x02\x1b\x43halk\\Common\\V1\\GPBMetadata\xea\x02\x11\x43halk::Common::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.common.v1.query_log_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.common.v1B\rQueryLogProtoP\001\242\002\003CCX\252\002\017Chalk.Common.V1\312\002\017Chalk\\Common\\V1\342\002\033Chalk\\Common\\V1\\GPBMetadata\352\002\021Chalk::Common::V1"
    _globals["_VERSIONEDQUERYNAME"]._options = None
    _globals["_VERSIONEDQUERYNAME"]._serialized_options = b"\030\001"
    _globals["_QUERYLOGFILTERS"]._options = None
    _globals["_QUERYLOGFILTERS"]._serialized_options = b"\030\001"
    _globals["_GETQUERYLOGENTRIESPAGETOKEN"]._options = None
    _globals["_GETQUERYLOGENTRIESPAGETOKEN"]._serialized_options = b"\030\001"
    _globals["_GETQUERYLOGENTRIESREQUEST"]._options = None
    _globals["_GETQUERYLOGENTRIESREQUEST"]._serialized_options = b"\030\001"
    _globals["_QUERYLOGENTRY"]._options = None
    _globals["_QUERYLOGENTRY"]._serialized_options = b"\030\001"
    _globals["_GETQUERYLOGENTRIESRESPONSE"]._options = None
    _globals["_GETQUERYLOGENTRIESRESPONSE"]._serialized_options = b"\030\001"
    _globals["_VERSIONEDQUERYNAME"]._serialized_start = 160
    _globals["_VERSIONEDQUERYNAME"]._serialized_end = 289
    _globals["_QUERYLOGFILTERS"]._serialized_start = 292
    _globals["_QUERYLOGFILTERS"]._serialized_end = 791
    _globals["_GETQUERYLOGENTRIESPAGETOKEN"]._serialized_start = 794
    _globals["_GETQUERYLOGENTRIESPAGETOKEN"]._serialized_end = 945
    _globals["_GETQUERYLOGENTRIESREQUEST"]._serialized_start = 948
    _globals["_GETQUERYLOGENTRIESREQUEST"]._serialized_end = 1366
    _globals["_QUERYLOGENTRY"]._serialized_start = 1369
    _globals["_QUERYLOGENTRY"]._serialized_end = 2176
    _globals["_GETQUERYLOGENTRIESRESPONSE"]._serialized_start = 2179
    _globals["_GETQUERYLOGENTRIESRESPONSE"]._serialized_end = 2309
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import uuid
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Dict, Mapping, Protocol, Sequence, Union

from chalk.client.models import (
    ChalkError,
    DatasetFilter,
    DatasetRevisionPreviewResponse,
    DatasetRevisionSummaryResponse,
    FeatureReference,
    FeatureResult,
    QueryMeta,
    QueryStatus,
)
from chalk.features import DataFrame
from chalk.features.resolver import ResolverProtocol
from chalk.integrations.catalogs.base_catalog import BaseCatalog

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl
    import pyarrow as pa


class DatasetPartition(Protocol):
    """Per-partition info for a dataset"""

    performance_summary: str | None
    """Performance information for computing this shard of the dataset"""


class DatasetRevision(Protocol):
    """Class wrapper around revisions for Datasets."""

    revision_id: uuid.UUID
    """UUID for the revision job."""

    creator_id: str
    """UUID for the creator of the job."""

    outputs: list[str]
    """Output features for the dataset revision."""

    givens_uri: str | None
    """Location of the givens stored for the dataset."""

    status: QueryStatus
    """Status of the revision job."""

    filters: DatasetFilter
    """Filters performed on the dataset."""

    num_partitions: int
    """Number of partitions for revision job."""

    output_uris: str
    """Location of the outputs stored for the dataset."""

    output_version: int
    """Storage version of the outputs."""

    num_bytes: int | None = None
    """Number of bytes of the output, updated upon success."""

    created_at: datetime | None = None
    """Timestamp for creation of revision job."""

    started_at: datetime | None = None
    """Timestamp for start of revision job."""

    terminated_at: datetime | None = None
    """Timestamp for end of revision job."""

    dataset_name: str | None = None
    """Name of revision, if given."""

    dataset_id: uuid.UUID | None = None
    """ID of revision, if name is given."""

    dashboard_url: str | None = None
    """url linking to relevant dashboard page"""

    environment: str

    num_computers: int
    """Number of computers this query ran on."""

    branch: str | None = None
    """Name of branch"""

    errors: list[ChalkError] | None = None

    partitions: list[DatasetPartition]

    metadata: Mapping[str, Any] | None = None

    @property
    def url(self) -> str | None:
        """url linking to relevant dashboard page"""
        return self.dashboard_url

    def to_polars(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pl.DataFrame:
        """Loads a `pl.DataFrame` containing the output. Use `.to_polars_lazyframe()` if you want
        a `LazyFrame` instead, which allows local filtering of datasets that are larger than memory.

        Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pl.LazyFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pl.LazyFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pl.DataFrame
            A `polars.DataFrame` materializing query output data.
        """
        ...

    def to_polars_lazyframe(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pl.LazyFrame:
        """Loads a `pl.LazyFrame` containing the output. This method is appropriate for working with larger-than-memory datasets.
        Use `.to_polars()` if you want a `DataFrame` instead.

        Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pl.LazyFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pl.LazyFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pl.LazyFrame
            A `pl.LazyFrame` materializing query output data.
        """
        ...

    def get_data_as_polars(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pl.LazyFrame:
        """Loads a `pl.LazyFrame` containing the output.

        Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pl.LazyFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pl.LazyFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pl.LazyFrame
            A `pl.LazyFrame` materializing query output data.
        """
        ...

    def get_data_as_pandas(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pd.DataFrame:
        """Loads a `pd.DataFrame` containing the output.

        Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pd.DataFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pd.DataFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pd.DataFrame
            A `pd.DataFrame` materializing query output data.
        """
        ...

    def get_data_as_dataframe(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> DataFrame:
        """Loads a Chalk `DataFrame` containing the output.

        Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `DataFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `DataFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data\
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        DataFrame
            A `DataFrame` materializing query output data.
        """
        ...

    def arrow_schema(
        self,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
    ) -> pa.Schema:
        """Returns the schema of the output data.

        Parameters
        ----------
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pa.Schema
            The schema of the output data.
        """
        ...

    def to_arrow(
        self,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pa.Table:
        """Loads a `pa.Table` from the raw parquet file outputs.

        Parameters
        ----------
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pa.Table
            A `pa.Table` materializing query output data.
        """
        ...

    def download_uris(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> Sequence[str]:
        """
        Returns a list of the output uris for the revision. Data
        will be stored in .parquet format. The URIs should be considered temporary,
        and will expire after a server-defined time period.
        """
        ...

    def summary(self) -> DatasetRevisionSummaryResponse:
        """
        Returns an object that loads the summary statistics of a dataset revision.
        The dataframe can be retrieved by calling `to_polars()` or `to_pandas()` on the return object.
        Data will be stored in .parquet format. The URIs should be considered temporary,
        and will expire after a server-defined time period.
        """
        ...

    def preview(self) -> DatasetRevisionPreviewResponse:
        """
        Returns an object that loads a preview of a dataset revision.
        The dataframe can be retrieved by calling `to_polars()` or `to_pandas()` on the return object.
        Data will be stored in .parquet format. The URIs should be considered temporary,
        and will expire after a server-defined time period.
        """
        ...

    def wait(
        self,
        timeout: float | timedelta | ellipsis | None = ...,
        show_progress: bool | ellipsis = ...,
    ) -> None:
        """
        Waits for an offline query job to complete.
        Raises if the query is unsuccessful, otherwise returns itself on success.

        Parameters
        ----------
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        show_progress
            Whether to show a progress bar. Defaults to True.
        """
        ...

    def download_data(
        self,
        path: str,
        output_id: bool = False,
        output_ts: Union[bool, str] = False,
        ignore_errors: bool = False,
        executor: ThreadPoolExecutor | None = None,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> None:
        """Downloads output files pertaining to the revision to given path.

        Datasets are stored in Chalk as sharded Parquet files. With this
        method, you can download those raw files into a directory for processing
        with other tools.

        Parameters
        ----------
        path
            A directory where the Parquet files from the dataset will be downloaded.
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `DataFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `DataFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        executor:
            The executor to use for parallelizing downloads. If None, the default executor will be used.
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        """
        ...

    def get_input_dataframe(
        self,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pl.LazyFrame:
        """Loads a `pl.LazyFrame` containing the inputs.

        Parameters
        ----------
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pl.LazyFrame
            A `pl.LazyFrame` materializing query input data.
        """
        ...

    def open_in_browser(self, return_url_only: bool = False) -> str:
        """Returns and opens a url that opens the offline query page in
        the Chalk dashboard. Must be logged in.

        Parameters
        ----------
        return_url_only
            If `True`, does not open url in browser. Default is `False`.

        Returns
        -------
        str
            A url redirecting to the Chalk dashboard.
        """
        ...

    def wait_for_completion(
        self,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
    ) -> None:
        """Waits for the revision job to complete.

        `ChalkClient.offline_query` returns a `DatasetRevision` instance immediately after
        submitting the revision job. This method can be used to wait for the
        revision job to complete.

        Once the revision job is complete, the `status` attribute of the
        `DatasetRevision` instance will be updated to reflect the status of the
        revision job.

        If the revision job was successful, you can then use methods such as
        `get_data_as_pandas()` without having to wait for the revision job to
        complete.

        Parameters
        ----------
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        """
        ...

    def ingest(self, store_online: bool = False, store_offline: bool = True) -> Dataset:
        """Saves this revision to Chalk's online and offline storage.

        Parameters
        ----------
        store_online
            Whether to store the revision in Chalk's online storage.
        store_offline
            Whether to store the revision in Chalk's offline storage.
        """
        ...

    def resolver_replay(
        self,
        resolver: ResolverProtocol,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> Union[pl.DataFrame, pl.LazyFrame, Mapping[str, pl.DataFrame], Mapping[str, pl.LazyFrame],]:
        """
        Downloads the resolver replay data for the given resolver in the revision, provided the revision had store_plan_stages
        enabled.

        The replay data is functionally similar to viewing the intermediate results on the plan explorer.

        If the resolver appears in only one stage of the plan, the resolver's replay data is returned directly.
        If the resolver instead appears in multiple stages of the plan, a mapping of the operation's ID to the replay data
        will be returned. If the resolver does not appear in the plan, an exception will be thrown.

        Parameters
        ----------
        resolver
            The resolver to download the replay data for, or its fqn.
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        """
        ...

    def set_metadata(
        self,
        metadata: Mapping[str, Any],
    ):
        """
        Set metadata for a dataset revision.

        Parameters
        ----------
        metadata
            The metadata (as a dict) that you want to set for a given revision—this will fully replace
            any metadata that has already been previously set.

        Examples
        --------
        >>> from chalk.client import ChalkClient, Dataset
        >>> dataset: Dataset = ChalkClient().get_dataset(dataset_name='my_dataset_name')
        >>> dataset.revisions[0].set_metadata(
        ...     {"metadata": "test"}
        ... )
        """


class Dataset(Protocol):
    """Wrapper around Offline Query results.

    Datasets are obtained by invoking `ChalkClient.offline_query()`.
    `Dataset` instances store important metadata and enable the retrieval of
    offline query outputs.

    Examples
    --------
    >>> from chalk.client import ChalkClient, Dataset
    >>> uids = [1, 2, 3, 4]
    >>> at = datetime.now(tz=timezone.utc)
    >>> dataset: Dataset = ChalkClient().offline_query(
    ...     input={
    ...         User.id: uids,
    ...     },
    ...     input_times=[at] * len(uids),
    ...     output=[
    ...         User.id,
    ...         User.fullname,
    ...         User.email,
    ...         User.name_email_match_score,
    ...     ],
    ...     dataset_name='my_dataset'
    ... )
    >>> df = dataset.get_data_as_pandas()
    >>> df.recompute(features=[User.fraud_score], branch="feature/testing")
    """

    is_finished: bool
    """Whether the most recent `DatasetRevision` is finished or still pending."""

    version: int
    """Storage version number of outputs."""

    revisions: Sequence[DatasetRevision]
    """A list of all `DatasetRevision` instances belonging to this dataset."""

    dataset_name: str | None
    """The unique name for this dataset, if given."""

    dataset_id: uuid.UUID | None
    """The unique UUID for this dataset."""

    errors: Sequence[ChalkError] | None
    """A list of errors in loading the dataset, if they exist."""

    def to_polars(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pl.DataFrame:
        """Loads a `pl.DataFrame` containing the output. Use `.to_polars_lazyframe()` if you want
        a `LazyFrame` instead, which allows local filtering of datasets that are larger than memory.

        Other Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pl.LazyFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pl.LazyFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a `TimeoutError`.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pl.DataFrame
            A `pl.DataFrame` materializing query output data.
        """
        ...

    async def to_polars_async(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pl.DataFrame:
        """Loads a `pl.DataFrame` containing the output. Use `.to_polars_lazyframe()` if you want
        a `LazyFrame` instead, which allows local filtering of datasets that are larger than memory.

        Other Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pl.LazyFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pl.LazyFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a `TimeoutError`.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pl.DataFrame
            A `pl.DataFrame` materializing query output data.
        """
        ...

    def to_arrow(
        self,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pa.Table:
        """Loads a `pa.Table` from the raw parquet file outputs.

        Other Parameters
        ----------
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a `TimeoutError`.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pa.Table
            A `pa.Table` materializing query output data.
        """
        ...

    def arrow_schema(
        self,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
    ) -> pa.Schema:
        """Returns the schema of the output data.

        Parameters
        ----------
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pa.Schema
            The schema of the output data.
        """
        ...

    def to_polars_lazyframe(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pl.LazyFrame:
        """Loads a `pl.LazyFrame` containing the output. This method is appropriate for working with larger-than-memory datasets.
        Use `.to_polars()` if you want a `DataFrame` instead.

        Other Parameters
        ----------------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pl.LazyFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pl.LazyFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pl.LazyFrame
            A `pl.LazyFrame` materializing query output data.
        """
        ...

    def get_data_as_polars(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pl.LazyFrame:
        """Loads a `pl.LazyFrame` containing the output.

        Other Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pl.LazyFrame`.
        output_ts
            Whether to return the timestamp feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pl.LazyFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pl.LazyFrame
            A `pl.LazyFrame` materializing query output data.
        """
        ...

    def get_data_as_pandas(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pd.DataFrame:
        """Loads a `pd.DataFrame` containing the output.

        Other Parameters
        ----------------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pd.DataFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pd.DataFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pd.DataFrame
            A `pd.DataFrame` materializing query output data.
        """
        ...

    def get_data_as_dataframe(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> DataFrame:
        """Loads a Chalk `DataFrame` containing the output.
        Requires the pertinent Chalk features to be accessible via import

        Other Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `DataFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `DataFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        DataFrame
            A `DataFrame` materializing query output data.
        """
        ...

    def to_pandas(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pd.DataFrame:
        """Loads a `pd.DataFrame` containing the output of the most recent revision.

        Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pd.DataFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pd.DataFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pd.DataFrame
            A `pd.DataFrame` materializing query output data.
        """
        ...

    def download_uris(
        self,
        output_id: bool = False,
        output_ts: bool | str = False,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> list[str]:
        """
        Returns a list of the output uris for the revision. Data
        will be stored in .parquet format. The URIs should be considered temporary,
        and will expire after a server-defined time period.

        Parameters
        ----------
        output_id
            Whether to return the primary key feature in a column
            named `"__chalk__.__id__"` in the resulting `pd.DataFrame`.
        output_ts
            Whether to return the input-time feature in a column
            named `"__chalk__.CHALK_TS"` in the resulting `pd.DataFrame`.
            If set to a non-empty `str`, used as the input-time column name.
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        """
        ...

    def wait(
        self,
        timeout: float | timedelta | ellipsis | None = ...,
        show_progress: bool | ellipsis = ...,
    ) -> Dataset:
        """
        Waits for an offline query job to complete. Returns a list of errors if unsuccessful, or None if successful.

        Parameters
        ----------
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        show_progress
            Whether to show a progress bar. Defaults to True.
        """
        ...

    def download_data(
        self,
        path: str,
        executor: ThreadPoolExecutor | None = None,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> None:
        """Downloads output files pertaining to the revision to the given path.

        Datasets are stored in Chalk as sharded Parquet files. With this
        method, you can download those raw files into a directory for processing
        with other tools.

        Parameters
        ----------
        path
            A directory where the Parquet files from the dataset will be downloaded.
        ignore_errors
            Whether to ignore query errors upon fetching data.
        executor
            An executor to use to download the data in parallel. If not specified, the
            default executor will be used.
        show_progress
            Whether to show a progress bar. Defaults to True..
        timeout
            How long to wait, in seconds, for job completion before raising a `TimeoutError`.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Examples
        --------
        >>> from chalk.client import ChalkClient, Dataset
        >>> from datetime import datetime, timezone
        >>> uids = [1, 2, 3, 4]
        >>> at = datetime.now(tz=timezone.utc)
        >>> dataset = ChalkClient().offline_query(
        ...     input={User.id: uids},
        ...     input_times=[at] * len(uids),
        ...     output=[
        ...         User.id,
        ...         User.fullname,
        ...         User.email,
        ...         User.name_email_match_score,
        ...     ],
        ...     dataset_name='my_dataset',
        ... )
        >>> dataset.download_data('my_directory')
        """
        ...

    def summary(self) -> DatasetRevisionSummaryResponse:
        """
        Returns an object that loads the summary statistics of a dataset revision.
        The dataframe can be retrieved by calling `to_polars()` or `to_pandas()` on the return object.
        Data will be stored in .parquet format. The URIs should be considered temporary,
        and will expire after a server-defined time period.
        """
        ...

    def preview(self) -> DatasetRevisionPreviewResponse:
        """
        Returns an object that loads a preview of a dataset revision.
        The dataframe can be retrieved by calling `to_polars()` or `to_pandas()` on the return object.
        Data will be stored in .parquet format. The URIs should be considered temporary,
        and will expire after a server-defined time period.
        """
        ...

    def get_input_dataframe(
        self,
        ignore_errors: bool = False,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> pl.LazyFrame:
        """Loads a `pl.LazyFrame` containing the inputs that were used to create the dataset.

        Parameters
        ----------
        ignore_errors
            Whether to ignore query errors upon fetching data
        show_progress
            Whether to show a progress bar. Defaults to True.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.

        Returns
        -------
        pl.LazyFrame
            A `pl.LazyFrame` materializing query input data.
        """
        ...

    def open_in_browser(self, return_url_only: bool = False) -> str:
        """Returns and opens a url that opens the offline query page
        in the Chalk dashboard. Must be logged in.

        Parameters
        ----------
        return_url_only
            If True, does not open url in browser. Default is False.

        Returns
        -------
        str
            A url redirecting to the Chalk dashboard.
        """
        ...

    def recompute(
        self,
        features: list[FeatureReference] | None = None,
        branch: str | None = None,
        wait: bool = True,
        show_progress: bool | ellipsis = ...,
        store_plan_stages: bool = False,
        correlation_id: str | None = None,
        explain: bool = False,
        tags: list[str] | None = None,
        required_resolver_tags: list[str] | None = None,
        planner_options: Mapping[str, Union[str, int, bool]] | None = None,
        run_asynchronously: bool = False,
        timeout: float | timedelta | None | ellipsis = ...,
    ) -> Dataset:
        """Creates a new revision of this `Dataset` by recomputing the specified features.

        Carries out the new computation on the branch specified when constructing the client.

        Parameters
        ----------
        features
            A list of specific features to recompute. Features that don't exist in the dataset will be added.
            Features that already exist in the dataset will be recomputed.
            If not provided, all the existing features in the dataset will be recomputed.
        branch
            If specified, Chalk will route your request to the relevant branch.
            If None, Chalk will route your request to a non-branch deployment.
            If not specified, Chalk will use the current client's branch info.
        show_progress
            If True, progress bars will be shown while recomputation is running.
            This flag will also be propagated to the methods of the resulting
            `Dataset`.
        correlation_id
            You can specify a correlation ID to be used in logs and web interfaces.
            This should be globally unique, i.e. a `uuid` or similar. Logs generated
            during the execution of your query will be tagged with this correlation id.
        store_plan_stages
            If True, the output of each of the query plan stages will be stored
            in S3/GCS. This will dramatically impact the performance of the query,
            so it should only be used for debugging.
            These files will be visible in the web dashboard's query detail view, and
            can be downloaded in full by clicking on a plan node in the query plan visualizer.
        tags
            The tags used to scope the resolvers.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        required_resolver_tags
            If specified, *all* required_resolver_tags must be present on a resolver for it to be
            considered eligible to execute.
            See https://docs.chalk.ai/docs/resolver-tags for more information.
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        run_asynchronously
            Boots a kubernetes job to run the queries in their own pods, separate from the engine and branch servers.
            This is useful for large datasets and jobs that require a long time to run.
            This must be specified as True to run this job asynchronously,
            even if the previous revision was run asynchronously.

        Raises
        ------
        ValueError
            If no branch was provided to the Chalk Client.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> dataset = ChalkClient(branch="data_science").offline_query(...)
        >>> df = dataset.get_data_as_polars()
        >>> # make changes to resolvers in your project
        >>> dataset.recompute()
        >>> new_df = dataset.get_data_as_polars() # receive newly computed data
        """
        ...

    def ingest(self, store_online: bool = False, store_offline: bool = True) -> Dataset:
        """
        Saves the latest revision of this dataset to Chalk's online and offline storage.

        Parameters
        ----------
        store_online
            Whether to store the revision in Chalk's online storage.
        store_offline
            Whether to store the revision in Chalk's offline storage.
        """
        ...

    def resolver_replay(
        self,
        resolver: ResolverProtocol,
        show_progress: bool | ellipsis = ...,
        timeout: float | timedelta | ellipsis | None = ...,
    ) -> Union[pl.DataFrame, pl.LazyFrame, Mapping[str, pl.DataFrame], Mapping[str, pl.LazyFrame],]:
        """
        Downloads the resolver replay data for the given resolver in the latest revision of the dataset.

        The replay data is functionally similar to viewing the intermediate results on the plan explorer.

        If the resolver appears in only one stage of the plan, the resolver's replay data is returned directly.
        If the resolver instead appears in multiple stages of the plan, a mapping of the operation's ID to the replay data
        will be returned. If the resolver does not appear in the plan, an exception will be thrown.

        Parameters
        ----------
        resolver
            The resolver to download the replay data for, or its fqn.
        show_progress
            Whether to show progress bars
        timeout
            How long to wait, in seconds, for job completion before raising a TimeoutError.
            Jobs will continue to run in the background if they take longer than this timeout.
            For no timeout, set to `None`. If no timeout is specified, the client's default
            timeout is used.
        """
        ...

    def write_to(self, destination: str, catalog: BaseCatalog | None = None) -> None:
        """
        Writes the dataset to a given destination.

        Parameters
        ----------
        destination
            The destination to write the dataset to.
        catalog
            The catalog to use for writing the dataset.
        """
        ...

    def set_metadata(
        self,
        metadata: Mapping[str, Any],
    ):
        """
        Set metadata for the latest dataset revision of the dataset.

        Parameters
        ----------
        metadata
            The metadata (as a dict) that you want to set for a given revision—this will fully replace
            any metadata that has already been previously set.

        Examples
        --------
        >>> from chalk.client import ChalkClient, Dataset
        >>> dataset: Dataset = ChalkClient().get_dataset(dataset_name='my_dataset_name')
        >>> dataset.set_metadata(
        ...     {"metadata": "test"}
        ... )
        """
        ...


class BulkQueryResponse(Protocol):
    scalars_df: pl.DataFrame | None
    groups_dfs: Dict[str, pl.DataFrame] | None
    errors: list[ChalkError] | None
    meta: QueryMeta | None

    def get_feature_value(self, feature: FeatureReference) -> Any:
        """Convenience method for accessing feature values from the data response.

        Parameters
        ----------
        feature
            The feature or its string representation.

        Returns
        -------
        Any
            The value of the feature.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> data = ChalkClient().query(...)
        >>> data.get_feature_value(User.name)
        "Katherine Johnson"
        >>> data.get_feature_value("user.name")
        "Katherine Johnson"
        """
        ...


class QueryBulkResponse(Protocol):
    responses: list[BulkQueryResponse]


class OnlineQueryResult(Protocol):
    data: list[FeatureResult]
    """The output features and any query metadata."""

    errors: list[ChalkError] | None
    """Errors encountered while running the resolvers.

    If no errors were encountered, this field is empty.
    """

    meta: QueryMeta | None
    """Metadata about the query execution.
    Only present if `include_meta=True` is passed to the relevant query method.
    """

    def get_feature(self, feature: FeatureReference) -> FeatureResult | None:
        """Convenience method for accessing feature result from the data response.

        Parameters
        ----------
        feature
            The feature or its string representation.

        Returns
        -------
        FeatureResult | None
            The `FeatureResult` for the feature, if it exists.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> data = ChalkClient().query(...)
        >>> data.get_feature(User.name).ts
        datetime.datetime(2023, 2, 5, 23, 25, 26, 427605)
        >>> data.get_feature("user.name").meta.cache_hit
        False
        """
        ...

    def get_feature_value(self, feature: FeatureReference) -> Any:
        """Convenience method for accessing feature values from the data response.

        Parameters
        ----------
        feature
            The feature or its string representation.

        Returns
        -------
        Any
            The value of the feature.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> data = ChalkClient().query(...)
        >>> data.get_feature_value(User.name)
        "Katherine Johnson"
        >>> data.get_feature_value("user.name")
        "Katherine Johnson"
        """
        ...

    def to_dict(self, prefix: bool = True) -> dict[str, Any]:
        """Converts the output features to a dictionary.
        Errors are not included in the dictionary.

        Returns
        -------
        dict[str, Any]
            A dictionary of the output features.

        Examples
        --------
        >>> from chalk.client import ChalkClient
        >>> result = ChalkClient().query(...)
        >>> result.to_dict()
        {
            "user.name": "Katherine Johnson",
            "user.email": "katherine@nasa.com"
        }
        """
        ...



from __future__ import annotations

import json
from typing import Any, Optional


class StreamSource:
    """Base class for all stream sources generated from `@stream`."""

    registry: "list[StreamSource]" = []
    name: Optional[str] = None

    def _config_to_json(self) -> Any:
        return self.config_to_json()  # for backcompat

    def config_to_json(self) -> str:
        raise NotImplementedError()

    def config_to_dict(self) -> dict:
        return json.loads(self.config_to_json())

    def _recreate_integration_variables(self) -> dict[str, str]:
        raise NotImplementedError()

    @property
    def streaming_type(self) -> str:
        """e.g. 'kafka' or 'kinesis' or 'pubsub'"""
        raise NotImplementedError()

    @property
    def dlq_name(self) -> str | None:
        """
        Identifier for the dead-letter queue (DLQ) for the stream.
        If not specified, failed messages will be dropped.
        Stream name for Kinesis, topic name for Kafka, subscription id for PubSub.
        """
        raise NotImplementedError()

    @property
    def stream_or_topic_name(self) -> str:
        """
        Identifier for the stream to consume.
        Stream name for Kinesis, topic name for Kafka, subscription id for PubSub
        """
        raise NotImplementedError()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.log_pb2 import (
    SearchLogEntriesRequest,
    SearchLogEntriesResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class LogSearchServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    SearchLogEntries: UnaryUnaryMultiCallable[
        SearchLogEntriesRequest,
        SearchLogEntriesResponse,
    ]

class LogSearchServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def SearchLogEntries(
        self,
        request: SearchLogEntriesRequest,
        context: ServicerContext,
    ) -> SearchLogEntriesResponse: ...

def add_LogSearchServiceServicer_to_server(servicer: LogSearchServiceServicer, server: Server) -> None: ...



import dataclasses
import json
import os
import re
import traceback
from pathlib import Path
from typing import List, Optional, Union

from chalk._lsp.error_builder import LSPErrorBuilder
from chalk._monitoring.Chart import Chart
from chalk._monitoring.gql_conversion import convert_chart
from chalk._version import __version__
from chalk.config.project_config import ProjectSettings, load_project_config
from chalk.features import FeatureSetBase
from chalk.features.resolver import RESOLVER_REGISTRY
from chalk.importer import CHALK_IMPORTER, FailedImport
from chalk.parsed._graph_validation import validate_graph
from chalk.parsed.duplicate_input_gql import (
    ChalkPYInfo,
    CreateChartGQL,
    EnvironmentSettingsGQL,
    FeatureClassGQL,
    FeatureSettings,
    LspGQL,
    MetadataSettings,
    ProjectSettingsGQL,
    PublishDiagnosticsParams,
    ResolverSettings,
    UpsertCronQueryGQL,
    UpsertGraphGQL,
    UpsertNamedQueryGQL,
    UpsertResolverGQL,
    UpsertSinkResolverGQL,
    UpsertSQLSourceGQL,
    UpsertStreamResolverGQL,
    ValidationSettings,
)
from chalk.parsed.json_conversions import convert_type_to_gql, gather_cdc_sources
from chalk.queries.named_query import NAMED_QUERY_REGISTRY
from chalk.queries.scheduled_query import CRON_QUERY_REGISTRY
from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.utils import paths
from chalk.utils.paths import get_directory_root


def _is_relative_to(x: Path, other: Path) -> bool:
    try:
        x.relative_to(other)
        return True
    except ValueError:
        return False


def project_settings_to_gql(config: ProjectSettings) -> ProjectSettingsGQL:
    def read_packages(filename: str) -> Optional[List[str]]:
        reqs = list()
        try:
            with open(filename) as f:
                for r in f.readlines():
                    cleaned = re.sub("#.*", "", r).rstrip("\n").strip()
                    if cleaned != "":
                        reqs.append(cleaned)
            return reqs
        except OSError:
            return None

    return ProjectSettingsGQL(
        project=config.project,
        environments=(
            None
            if config.environments is None
            else [
                EnvironmentSettingsGQL(
                    id=i,
                    runtime=e.runtime,
                    requirements=e.requirements,
                    dockerfile=e.dockerfile,
                    requiresPackages=(
                        None
                        if e.requirements is None
                        else read_packages(
                            os.path.join(
                                os.path.dirname(config.local_path),
                                e.requirements,
                            )
                        )
                    ),
                    platformVersion=e.platform_version,
                )
                for i, e in config.environments.items()
            ]
        ),
        validation=(
            ValidationSettings(
                feature=(
                    FeatureSettings(
                        metadata=(
                            [
                                MetadataSettings(name=m.name, missing=m.missing)
                                for m in config.validation.feature.metadata
                            ]
                            if config.validation.feature.metadata
                            else None
                        )
                    )
                    if config.validation.feature
                    else None
                ),
                resolver=(
                    ResolverSettings(
                        metadata=(
                            [
                                MetadataSettings(name=m.name, missing=m.missing)
                                for m in config.validation.resolver.metadata
                            ]
                            if config.validation.resolver.metadata
                            else None
                        )
                    )
                    if config.validation.resolver
                    else None
                ),
            )
            if config.validation
            else None
        ),
    )


def get_registered_types(scope_to: Path, failed: List[FailedImport]) -> UpsertGraphGQL:
    features = []
    feature_classes: list[FeatureClassGQL] = []
    for x in FeatureSetBase.registry.values():
        cls_path = paths.get_classpath(x)
        if cls_path is not None and _is_relative_to(cls_path, scope_to):
            try:
                feature_classes.append(convert_type_to_gql(x))
            except Exception as e:
                failed.append(build_failed_import(e, f"feature class '{x.namespace}'"))
                continue
            try:
                for feature in x.features:
                    if not feature.is_autogenerated and not feature.no_display:
                        try:
                            converted = convert_type_to_gql(feature)
                            features.append(converted)
                        except Exception as e:
                            if not LSPErrorBuilder.promote_exception(e):
                                failed.append(build_failed_import(e, f"feature '{feature.fqn}'"))
            except Exception as e:
                if not LSPErrorBuilder.promote_exception(e):
                    failed.append(build_failed_import(e, f"feature class '{x.namespace}'"))
    project_path = get_directory_root()
    path_prefix_to_remove = project_path.as_posix() if project_path is not None else None

    stream_resolvers: list[UpsertStreamResolverGQL] = []
    for stream in RESOLVER_REGISTRY.get_stream_resolvers():
        if _is_relative_to(Path(stream.filename), scope_to):
            try:
                stream_resolvers.append(convert_type_to_gql(stream, path_prefix=path_prefix_to_remove))
            except Exception as e:
                failed.append(build_failed_import(e, f"streaming resolver '{stream.fqn}'"))

    resolvers: list[UpsertResolverGQL] = []
    for resolver in RESOLVER_REGISTRY.get_online_and_offline_resolvers():
        if _is_relative_to(Path(resolver.filename), scope_to):
            try:
                resolvers.append(convert_type_to_gql(resolver, path_prefix=path_prefix_to_remove))
            except Exception as e:
                if not LSPErrorBuilder.promote_exception(e):
                    failed.append(build_failed_import(e, f"resolver '{resolver.fqn}'"))

    sink_resolvers: list[UpsertSinkResolverGQL] = []
    for sink in RESOLVER_REGISTRY.get_sink_resolvers():
        if _is_relative_to(Path(sink.filename), scope_to):
            try:
                sink_resolvers.append(convert_type_to_gql(sink, path_prefix=path_prefix_to_remove))
            except Exception as e:
                failed.append(build_failed_import(e, f"sink resolver '{sink.fqn}'"))

    cron_queries: list[UpsertCronQueryGQL] = []
    for cron_query in CRON_QUERY_REGISTRY.values():
        if _is_relative_to(Path(cron_query.filename), scope_to):
            try:
                cron_queries.append(convert_type_to_gql(cron_query, path_prefix=path_prefix_to_remove))
            except Exception as e:
                failed.append(build_failed_import(e, f"cron query '{cron_query.name}'"))

    named_queries: list[UpsertNamedQueryGQL] = []
    for named_query in NAMED_QUERY_REGISTRY.values():
        named_query.input
        named_query.output
        if named_query.filename and _is_relative_to(Path(named_query.filename), scope_to):
            try:
                named_queries.append(convert_type_to_gql(named_query, path_prefix=path_prefix_to_remove))
            except Exception as e:
                failed.append(build_failed_import(e, f"named query '{named_query.name}'"))

    charts: list[CreateChartGQL] = []
    for chart in Chart.registry:
        try:
            charts.append(convert_chart(chart))
        except Exception as e:
            failed.append(build_failed_import(e, f"chart ' {chart.name}'"))

    config = load_project_config()
    if config is not None:
        config = project_settings_to_gql(config)

    graph = UpsertGraphGQL(
        streams=stream_resolvers,
        sinks=sink_resolvers,
        features=features,
        config=config,
        failed=failed,
        resolvers=resolvers,
        cronQueries=cron_queries,
        namedQueries=named_queries,
        charts=charts,
        chalkpy=ChalkPYInfo(
            version=__version__,
        ),
        cdcSources=gather_cdc_sources(),
        sqlSources=[
            UpsertSQLSourceGQL(
                name=source.name,
                kind=source.kind,
            )
            for source in BaseSQLSource.registry
        ],
        featureClasses=feature_classes,
    )
    errors = validate_graph(graph)
    graph.errors = errors

    # Validating the graph produces errors, so do this part last.
    lsp = get_lsp_gql()
    if LSPErrorBuilder.lsp:
        try:
            lsp.diagnostics = CHALK_IMPORTER.supplement_diagnostics(failed_imports=failed, diagnostics=lsp.diagnostics)
        except Exception as e:
            if len(lsp.diagnostics) == 0 and len(failed) == 0:
                raise e

    graph.lsp = lsp

    return graph


def get_lsp_gql() -> LspGQL:
    return LspGQL(
        diagnostics=[
            PublishDiagnosticsParams(
                uri=filename,
                diagnostics=diagnostics,
            )
            for filename, diagnostics in LSPErrorBuilder.all_errors.items()
        ],
        actions=LSPErrorBuilder.all_edits,
    )


def build_failed_import(error: Union[Exception, str], description: str) -> FailedImport:
    try:
        formatted_tb = error if isinstance(error, str) else "\n".join(traceback.format_exception(error))
    except:
        formatted_tb = (
            error
            if isinstance(error, str)
            else "\n".join(
                traceback.format_exception(
                    type(error),
                    value=error,
                    tb=None,
                )
            )
        )

    return FailedImport(
        filename="",
        module="",
        traceback=f"EXCEPTION in parsing {description}:\n{formatted_tb}",
    )


def get_registered_types_as_json(scope_to: Path, failed: List[FailedImport], indent: int = 2) -> str:
    return json.dumps(
        dataclasses.asdict(get_registered_types(scope_to, failed)),
        indent=indent,
    )



class GasLimit:
    """
    A way to limit recursive functions.
    """

    def __init__(self, remaining_gas: int, out_of_gas_error: Exception):
        self.remaining_gas = remaining_gas
        self.out_of_gas_error = out_of_gas_error
        super().__init__()

    def consume_gas(self, amount: int = 1) -> None:
        self.remaining_gas -= amount
        if self.remaining_gas <= 0:
            raise self.out_of_gas_error


class OutOfGasError(Exception):
    """A generic sentinel exception for running out of gas."""

    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/engine/v1/plan_server.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.common.v1 import online_query_pb2 as chalk_dot_common_dot_v1_dot_online__query__pb2
from chalk._gen.chalk.engine.v1 import plan_pb2 as chalk_dot_engine_dot_v1_dot_plan__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n!chalk/engine/v1/plan_server.proto\x12\x0f\x63halk.engine.v1\x1a"chalk/common/v1/online_query.proto\x1a\x1a\x63halk/engine/v1/plan.proto"g\n\x0eGetPlanRequest\x12U\n\x14online_query_request\x18\x01 \x01(\x0b\x32#.chalk.common.v1.OnlineQueryRequestR\x12onlineQueryRequest"<\n\x0fGetPlanResponse\x12)\n\x04plan\x18\x01 \x01(\x0b\x32\x15.chalk.engine.v1.PlanR\x04plan"l\n\x13\x45xecuteQueryRequest\x12U\n\x14online_query_request\x18\x01 \x01(\x0b\x32#.chalk.common.v1.OnlineQueryRequestR\x12onlineQueryRequest"p\n\x14\x45xecuteQueryResponse\x12X\n\x15online_query_response\x18\x01 \x01(\x0b\x32$.chalk.common.v1.OnlineQueryResponseR\x13onlineQueryResponse2\xbc\x01\n\x0bPlanService\x12N\n\x07GetPlan\x12\x1f.chalk.engine.v1.GetPlanRequest\x1a .chalk.engine.v1.GetPlanResponse"\x00\x12]\n\x0c\x45xecuteQuery\x12$.chalk.engine.v1.ExecuteQueryRequest\x1a%.chalk.engine.v1.ExecuteQueryResponse"\x00\x42\x84\x01\n\x13\x63om.chalk.engine.v1B\x0fPlanServerProtoP\x01\xa2\x02\x03\x43\x45X\xaa\x02\x0f\x43halk.Engine.V1\xca\x02\x0f\x43halk\\Engine\\V1\xe2\x02\x1b\x43halk\\Engine\\V1\\GPBMetadata\xea\x02\x11\x43halk::Engine::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.engine.v1.plan_server_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.engine.v1B\017PlanServerProtoP\001\242\002\003CEX\252\002\017Chalk.Engine.V1\312\002\017Chalk\\Engine\\V1\342\002\033Chalk\\Engine\\V1\\GPBMetadata\352\002\021Chalk::Engine::V1"
    _globals["_GETPLANREQUEST"]._serialized_start = 118
    _globals["_GETPLANREQUEST"]._serialized_end = 221
    _globals["_GETPLANRESPONSE"]._serialized_start = 223
    _globals["_GETPLANRESPONSE"]._serialized_end = 283
    _globals["_EXECUTEQUERYREQUEST"]._serialized_start = 285
    _globals["_EXECUTEQUERYREQUEST"]._serialized_end = 393
    _globals["_EXECUTEQUERYRESPONSE"]._serialized_start = 395
    _globals["_EXECUTEQUERYRESPONSE"]._serialized_end = 507
    _globals["_PLANSERVICE"]._serialized_start = 510
    _globals["_PLANSERVICE"]._serialized_end = 698
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import ast
import dataclasses
import difflib
import inspect
import textwrap
from collections import defaultdict
from typing import TYPE_CHECKING, Callable, Dict, List, Mapping, NoReturn, Optional, Type, overload

from executing import Source

from chalk._lsp._class_finder import get_function_ast
from chalk._lsp.finders import (
    get_annotation_range,
    get_class_definition_range,
    get_comment_range,
    get_decorator_kwarg_value_range,
    get_feature_range,
    get_full_comment_range,
    get_full_range,
    get_function_arg_annotations,
    get_function_arg_values,
    get_function_decorator_arg_by_name,
    get_function_decorator_range,
    get_function_name,
    get_function_return_annotation,
    get_function_return_statement,
    get_key_from_dict_node,
    get_missing_return_annotation,
    get_property_range,
    get_property_value_call_range,
    get_property_value_range,
    get_sql_range,
    get_value_from_dict_node,
    get_variable_range,
    node_to_range,
)
from chalk.parsed.duplicate_input_gql import (
    CodeActionGQL,
    CodeDescriptionGQL,
    DiagnosticGQL,
    DiagnosticRelatedInformationGQL,
    DiagnosticSeverityGQL,
    LocationGQL,
    PositionGQL,
    RangeGQL,
)
from chalk.utils.collections import OrderedSet
from chalk.utils.source_parsing import should_skip_source_code_parsing
from chalk.utils.string import oxford_comma_list

if TYPE_CHECKING:
    import types

    from sqlglot.expressions import Select, Union

    from chalk.features import FeatureWrapper


class DiagnosticBuilder:
    def __init__(
        self,
        severity: DiagnosticSeverityGQL,
        message: str,
        uri: str,
        range: RangeGQL,
        label: str,
        code: str,
        code_href: str | None,
    ):
        super().__init__()
        self.uri = uri
        self.diagnostic = DiagnosticGQL(
            range=range,
            message=message,
            severity=severity,
            code=code,
            codeDescription=CodeDescriptionGQL(href=code_href) if code_href is not None else None,
            relatedInformation=[
                DiagnosticRelatedInformationGQL(
                    location=LocationGQL(uri=uri, range=range),
                    message=label,
                )
            ],
        )

    def with_range(
        self,
        range: RangeGQL | ast.AST | None,
        label: str,
    ) -> DiagnosticBuilder:
        if isinstance(range, ast.AST):
            range = node_to_range(range)
        if range is None:
            return self

        assert self.diagnostic.relatedInformation is not None
        self.diagnostic.relatedInformation.append(
            DiagnosticRelatedInformationGQL(
                location=LocationGQL(
                    uri=self.uri,
                    range=range,
                ),
                message=label,
            )
        )
        return self


_dummy_builder = DiagnosticBuilder(
    severity=DiagnosticSeverityGQL.Error,
    message="",
    uri="",
    range=RangeGQL(
        start=PositionGQL(line=0, character=0),
        end=PositionGQL(line=0, character=0),
    ),
    label="",
    code="",
    code_href=None,
)


class LSPErrorBuilder:
    lsp: bool = False
    """This should ONLY be True if we're running `chalk export`.
    DO NOT SET THIS TO TRUE IN ANY OTHER CONTEXT.
    Talk to Elliot if you think you need to set this to True."""

    all_errors: Mapping[str, list[DiagnosticGQL]] = defaultdict(list)
    all_edits: list[CodeActionGQL] = []

    _exception_map: dict[int, tuple[str, DiagnosticGQL]] = {}
    _strong_refs: dict[int, Exception] = {}
    """Maintain exception_map's keys `id(exception)`.
    This could be done better with weakrefs, but you
    cant naively use a weakref.WeakKeyDictionary because
    we can't depend on the __eq__ method of the exception
    object."""

    _node_map: dict[tuple[FeatureWrapper, str], tuple[ast.AST, types.FrameType]] = {}

    @classmethod
    def has_errors(cls):
        return cls.lsp and len(cls.all_errors) > 0

    @classmethod
    def save_node(cls, wrapper: FeatureWrapper, item: str):
        frame = inspect.currentframe()
        if frame is None:
            return
        i = 0
        while i < 2 and frame is not None:
            frame = frame.f_back
            i += 1
        if frame is None:
            return
        node_map_key = (wrapper, item)
        if node_map_key not in cls._node_map:
            try:
                node = Source.executing(frame).node
            except Exception:
                return
            if node is not None:  # pyright: ignore[reportUnnecessaryComparison]
                try:
                    cls._node_map[node_map_key] = (node, frame)
                except:
                    pass

    @classmethod
    def get_node(cls, wrapper: FeatureWrapper, item: str) -> tuple[ast.AST, types.FrameType] | None:
        return cls._node_map.get((wrapper, item))

    @classmethod
    def save_exception(cls, e: Exception, uri: str, diagnostic: DiagnosticGQL):
        """Save an exception to be promoted to a diagnostic later.
        Some exceptions are handled (e.g. hasattr(...) handles AttributeError)
        and should not become diagnostics unless the error isn't handled."""
        cls._exception_map[id(e)] = (uri, diagnostic)
        cls._strong_refs[id(e)] = e

    @classmethod
    def promote_exception(cls, e: Exception) -> bool:
        """Promote a previously saved exception to a diagnostic.
        Returns whether the exception was promoted."""
        if id(e) in cls._exception_map:
            uri, diagnostic = cls._exception_map[id(e)]
            cls.all_errors[uri].append(diagnostic)
            del cls._exception_map[id(e)]
            del cls._strong_refs[id(e)]
            return True

        return False


class FeatureClassErrorBuilder:
    def __init__(self, uri: str, namespace: str, node: ast.ClassDef | None):
        super().__init__()
        self.uri = uri
        self.diagnostics: List[DiagnosticGQL] = []
        self.namespace = namespace
        self.node = node
        self.error_cache: OrderedSet[tuple[str, RangeGQL | ast.AST, str]] = OrderedSet()

    def property_range(self, feature_name: str) -> ast.AST | None:
        if self.node is None:
            return None

        return get_property_range(cls=self.node, name=feature_name)

    def annotation_range(self, feature_name: str) -> ast.AST | None:
        if self.node is None:
            return None

        return get_annotation_range(cls=self.node, name=feature_name)

    def property_value_range(self, feature_name: str) -> ast.AST | None:
        if self.node is None:
            return None

        return get_property_value_range(cls=self.node, name=feature_name)

    def property_value_kwarg_range(self, feature_name: str, kwarg: str) -> ast.AST | None:
        if self.node is None:
            return None

        return get_property_value_call_range(cls=self.node, name=feature_name, kwarg=kwarg)

    def decorator_kwarg_value_range(self, kwarg: str) -> ast.AST | None:
        if self.node is None:
            return None

        return get_decorator_kwarg_value_range(cls=self.node, kwarg=kwarg)

    def class_definition_range(self) -> RangeGQL | None:
        if self.node is None:
            return None

        return get_class_definition_range(cls=self.node, filename=self.uri)

    def invalid_attribute(
        self,
        root_feature_str: str,
        root_is_feature_class: bool,
        item: str,
        candidates: List[str],
        back: int,
        saved_frame: FeatureWrapper | None = None,
    ):
        back = back + 1
        message = (
            f"Invalid attribute '{item}' on feature {'class ' if root_is_feature_class else ''}"
            + f"'{root_feature_str}'."
        )
        if not LSPErrorBuilder.lsp:
            # Short circuit if we're not in an LSP context. What follows is expensive.
            raise AttributeError(message)

        if saved_frame is not None:
            saved_node_and_frame = LSPErrorBuilder.get_node(saved_frame, item)
            if saved_node_and_frame is None:
                raise AttributeError(message)
            node, frame = saved_node_and_frame
        else:
            frame: Optional[types.FrameType] = inspect.currentframe()
            i = 0
            while i < back and frame is not None:
                frame = frame.f_back
                i += 1

            if frame is None or i != back:
                raise AttributeError(message)

            try:
                node = Source.executing(frame).node
            except Exception:
                raise AttributeError(message)

        uri = frame.f_locals.get("__file__")
        if isinstance(node, ast.Attribute):
            if node.end_lineno is None or node.end_col_offset is None:
                raise AttributeError(message)
            node = RangeGQL(
                start=PositionGQL(
                    line=node.end_lineno,
                    character=node.end_col_offset - len(node.attr),
                ),
                end=PositionGQL(
                    line=node.end_lineno,
                    character=node.end_col_offset,
                ),
            )

        candidates = [f"'{c}'" for c in candidates if not c.startswith("_")]
        if len(candidates) > 0:
            all_scores = [
                (
                    difflib.SequenceMatcher(a=item, b=candidate).quick_ratio(),
                    candidate,
                )
                for candidate in candidates
            ]
            all_scores.sort(key=lambda x: -x[0])

            if len(candidates) > 5:
                prefix = "The closest options are"
                candidates = [c for (_, c) in all_scores[:5]]
            elif len(candidates) == 1:
                prefix = "The only valid option is"
            else:
                prefix = "Valid options are"

            message += f" {prefix} {oxford_comma_list(candidates)}."

        self.add_diagnostic(
            message=message,
            range=node,
            label="Invalid attribute",
            code="55",
            raise_error=AttributeError,
            uri=uri,
        )

    @overload
    def add_diagnostic(
        self,
        message: str,
        label: str,
        code: str,
        *,
        range: RangeGQL | ast.AST | None,
        code_href: str | None = None,
        severity: DiagnosticSeverityGQL = DiagnosticSeverityGQL.Error,
        raise_error: None = ...,
        uri: str | None = ...,
    ) -> DiagnosticBuilder:
        ...

    @overload
    def add_diagnostic(
        self,
        message: str,
        label: str,
        code: str,
        *,
        range: RangeGQL | ast.AST | None,
        code_href: str | None = None,
        severity: DiagnosticSeverityGQL = DiagnosticSeverityGQL.Error,
        raise_error: Type[Exception],
        uri: str | None = ...,
    ) -> NoReturn:
        ...

    def add_diagnostic(
        self,
        message: str,
        label: str,
        code: str,
        range: RangeGQL | ast.AST | None,
        code_href: str | None = None,
        severity: DiagnosticSeverityGQL = DiagnosticSeverityGQL.Error,
        raise_error: Type[Exception] | None = None,
        uri: str | None = None,
    ) -> DiagnosticBuilder:
        uri = self.uri if uri is None else uri
        if not LSPErrorBuilder.lsp:
            if raise_error is not None:
                raise raise_error(message)
            return _dummy_builder
        default_error = TypeError
        if range is None:
            raise raise_error(message) if raise_error else default_error(message)
        if isinstance(range, ast.AST):
            range = node_to_range(range)
            if range is None:
                raise raise_error(message) if raise_error else default_error(message)

        builder = DiagnosticBuilder(
            severity=severity,
            message=message,
            uri=uri,
            range=range,
            label=label,
            code=code,
            code_href=code_href,
        )

        error = None if raise_error is None else raise_error(message)
        if error is None:
            if (message, range, uri) not in self.error_cache:
                self.diagnostics.append(builder.diagnostic)
                LSPErrorBuilder.all_errors[uri].append(builder.diagnostic)
                self.error_cache.add((message, range, uri))
        else:
            LSPErrorBuilder.save_exception(error, uri, builder.diagnostic)
            raise error

        return builder


class ResolverErrorBuilder:
    def __init__(
        self,
        fn: Callable | None,
    ):
        super().__init__()
        self._fn = fn
        self.diagnostics: List[DiagnosticGQL] = []
        self._uri = ...
        self._node = ...

    @property
    def uri(self):
        if self._uri is ...:
            self._load_node_and_uri()
        assert self._uri is not ...
        return self._uri

    @property
    def node(self):
        if self._node is ...:
            self._load_node_and_uri()
        assert self._node is not ...
        return self._node

    def _load_node_and_uri(self):
        """Lazily loading the node and uri on first use, because parsing the source is slow. If there are no errors, then no need to parse it"""
        source_info: Optional[FunctionSource] = None
        if not should_skip_source_code_parsing() and self._fn is not None:
            try:
                filename = inspect.getfile(self._fn)
                resolver_source = inspect.getsource(self._fn)
                dedent_source = resolver_source and textwrap.dedent(resolver_source)
                try:
                    tree = get_function_ast(self._fn)
                except:
                    tree = None
                source_info = FunctionSource(
                    filename=filename,
                    source=resolver_source,
                    dedent_source=dedent_source,
                    tree=tree,
                )
            except:
                pass

        self._uri = source_info.filename if source_info is not None else "__main__"
        self._node = source_info and source_info.tree

    @overload
    def add_diagnostic(
        self,
        message: str,
        label: str,
        code: str,
        range: RangeGQL | ast.AST | None,
        code_href: str | None = None,
        severity: DiagnosticSeverityGQL = DiagnosticSeverityGQL.Error,
        raise_error: None = ...,
        uri: str | None = ...,
    ) -> DiagnosticBuilder:
        ...

    @overload
    def add_diagnostic(
        self,
        message: str,
        label: str,
        code: str,
        *,
        range: RangeGQL | ast.AST | None,
        code_href: str | None = None,
        severity: DiagnosticSeverityGQL = DiagnosticSeverityGQL.Error,
        raise_error: Type[Exception] = ...,
        uri: str | None = ...,
    ) -> NoReturn:
        ...

    def add_diagnostic(
        self,
        message: str,
        label: str,
        code: str,
        range: RangeGQL | ast.AST | None,
        code_href: str | None = None,
        severity: DiagnosticSeverityGQL = DiagnosticSeverityGQL.Error,
        raise_error: Type[Exception] | None = None,
        uri: str | None = None,
    ) -> DiagnosticBuilder:
        """

        :param message: longform description of error with names of attributes, etc.
        :param label: shortform category of error
        :param code: unique identifier of error kind
        :param range: line number + offset of start and end of text with error
        :param code_href: code_href: link to doc
        :param severity: is it an error? a warning?
        :param raise_error: if we cannot proceed, raise with this error kind and the message.
        :param uri: filepath
        :return:
        """
        if not LSPErrorBuilder.lsp:
            if raise_error is not None:
                raise raise_error(message)
            return _dummy_builder
        uri = self.uri if uri is None else uri
        default_error = TypeError
        if range is None:
            raise raise_error(message) if raise_error else default_error(message)

        if isinstance(range, ast.AST):
            range = node_to_range(range)
            if range is None:
                raise raise_error(message) if raise_error else default_error(message)

        builder = DiagnosticBuilder(
            severity=severity,
            message=message,
            uri=uri,
            range=range,
            label=label,
            code=code,
            code_href=code_href,
        )

        error = None if raise_error is None else raise_error(message)
        if error is None:
            self.diagnostics.append(builder.diagnostic)
            LSPErrorBuilder.all_errors[uri].append(builder.diagnostic)
        else:
            LSPErrorBuilder.save_exception(error, uri, builder.diagnostic)
            raise error
        return builder

    def function_decorator(self) -> ast.AST | None:
        if self.node is None:
            return None

        return get_function_decorator_range(node=self.node)

    def function_decorator_arg_by_name(self, name: str) -> ast.AST | None:
        if self.node is None:
            return None

        return get_function_decorator_arg_by_name(node=self.node, name=name)

    def function_decorator_key_from_dict(self, decorator_field: str, arg_name: str) -> ast.AST | None:
        if self.node is None:
            return None
        decorator_arg = get_function_decorator_arg_by_name(node=self.node, name=decorator_field)
        if not isinstance(decorator_arg, ast.Dict):
            return decorator_arg
        return get_key_from_dict_node(decorator_arg, arg_name) or decorator_arg

    def function_decorator_value_from_dict(self, decorator_field: str, arg_name: str) -> ast.AST | None:
        if self.node is None:
            return None
        decorator_arg = get_function_decorator_arg_by_name(node=self.node, name=decorator_field)
        if not isinstance(decorator_arg, ast.Dict):
            return decorator_arg
        return get_value_from_dict_node(decorator_arg, arg_name) or decorator_arg

    def function_arg_values(self) -> Dict[str, ast.AST | None]:
        if self.node is None:
            return {}

        return get_function_arg_values(node=self.node)

    def function_arg_value_by_name(self, name: str) -> ast.AST | None:
        return self.function_arg_values().get(name)

    def function_arg_value_by_index(self, index: int) -> ast.AST | RangeGQL | None:
        if self.node is None:
            return None

        if len(self.node.args.args) == 0:
            return get_function_name(self.node, self.uri)
        if index < len(self.node.args.args):
            return self.node.args.args[index]
        return None

    def function_arg_annotations(self) -> Dict[str, ast.AST | None]:
        if self.node is None:
            return {}

        return get_function_arg_annotations(node=self.node)

    def function_arg_annotation_by_name(self, name: str) -> ast.AST | None:
        return self.function_arg_annotations().get(name)

    def function_arg_annotation_by_index(self, index: int) -> ast.AST | None:
        if self.node is None:
            return None

        if index < len(self.node.args.args):
            return self.node.args.args[index].annotation
        return None

    def function_return_annotation(self) -> ast.AST | RangeGQL | None:
        if self.node is None:
            return None

        node_or_none = get_function_return_annotation(node=self.node)
        return node_or_none or get_missing_return_annotation(self.node, self.uri)

    def function_return_statements(self) -> List[ast.AST | None]:
        if self.node is None:
            return []

        return get_function_return_statement(node=self.node)

    def function_name(self) -> RangeGQL | None:
        if self.node is None:
            return None

        return get_function_name(self.node, self.uri)

    def string_in_node(self, node: ast.AST, string: str, text: list[str]) -> RangeGQL | None:
        start_line = range_or_node_to_start_line(node)
        end_line = range_or_node_to_end_line(node)
        if start_line is None or end_line is None:
            return None
        for i, line in enumerate(text):
            if i < start_line:
                continue
            if i > end_line:
                return None
            if i == start_line:
                start_char = range_or_node_to_start_char(node)
                if start_char is None:
                    return None
            else:
                start_char = 0
            if i == end_line:
                end_char = range_or_node_to_end_char(node)
                if end_char is None:
                    return None
            else:
                end_char = len(line)
            starting_index = line.find(string, start_char, end_char)
            if starting_index != -1:
                return RangeGQL(
                    start=PositionGQL(
                        line=i + 1,
                        character=starting_index,
                    ),
                    end=PositionGQL(line=i + 1, character=starting_index + len(string)),
                )
        return None


@dataclasses.dataclass
class FunctionSource:
    filename: str
    source: str | None
    dedent_source: str | None
    tree: ast.FunctionDef | ast.AsyncFunctionDef | None


def get_resolver_error_builder(fn: Callable) -> ResolverErrorBuilder:
    error_builder = ResolverErrorBuilder(fn=fn)
    return error_builder


class SQLFileResolverErrorBuilder:
    def __init__(self, uri: str, sql_string: str, has_import_errors: bool):
        super().__init__()
        self.uri = uri
        self.has_import_errors = has_import_errors
        self.diagnostics: List[DiagnosticGQL] = []
        self.sql_string = sql_string
        self.sql_lines = sql_string.splitlines()

    def add_diagnostic(
        self,
        message: str,
        label: str,
        code: str,
        range: RangeGQL | ast.AST | None,
        code_href: str | None = None,
        severity: DiagnosticSeverityGQL = DiagnosticSeverityGQL.Error,
        raise_error: Type[Exception] | None = None,
        uri: str | None = None,
    ) -> DiagnosticBuilder:
        """

        :param message: longform description of error with names of attributes, etc.
        :param label: shortform category of error
        :param code: unique identifier of error kind
        :param range: line number + offset of start and end of text with error
        :param code_href: code_href: link to doc
        :param severity: is it an error? a warning?
        :param raise_error: if we cannot proceed, raise with this error kind and the message.
        :param uri: filepath
        :return:
        """
        if self.has_import_errors:
            # pass: we don't need lsp
            return _dummy_builder
        if not LSPErrorBuilder.lsp:
            if raise_error is not None:
                raise raise_error(message)
            return _dummy_builder
        default_error = TypeError
        if range is None:
            raise raise_error(message) if raise_error else default_error(message)
        uri = self.uri if uri is None else uri

        if isinstance(range, ast.AST):
            range = node_to_range(range)
            if range is None:
                raise raise_error(message) if raise_error else default_error(message)

        builder = DiagnosticBuilder(
            severity=severity,
            message=message,
            uri=uri,
            range=range,
            label=label,
            code=code,
            code_href=code_href,
        )

        error = None if raise_error is None else raise_error(message)
        if error is None:
            self.diagnostics.append(builder.diagnostic)
            LSPErrorBuilder.all_errors[uri].append(builder.diagnostic)
        else:
            LSPErrorBuilder.save_exception(error, uri, builder.diagnostic)
            raise error

        return builder

    def comment_range_by_key(self, name: str) -> RangeGQL | None:
        return get_comment_range(self.sql_lines, name)

    def full_comment_range(self) -> RangeGQL | None:
        return get_full_comment_range(self.sql_lines)

    def variable_range_by_name(self, name: str) -> RangeGQL | None:
        return get_variable_range(self.sql_lines, name)

    def value_range_by_name(self, glot: Select | Union, name: str) -> RangeGQL | None:
        return get_feature_range(self.sql_lines, glot, name)

    def custom_range(self, line_no: int, col: int, length: int | None = None) -> RangeGQL:
        length = length or 1
        return RangeGQL(
            start=PositionGQL(
                line=line_no,
                character=col,
            ),
            end=PositionGQL(
                line=line_no,
                character=col + length,
            ),
        )

    def full_range(self) -> RangeGQL:
        return get_full_range(self.sql_lines)

    def sql_range(self) -> RangeGQL | None:
        return get_sql_range(self.sql_lines)

    def add_diagnostic_with_spellcheck(
        self,
        spellcheck_item: str,
        spellcheck_candidates: List[str],
        message: str,
        label: str,
        code: str,
        range: RangeGQL | ast.AST | None,
        code_href: str | None = None,
        severity: DiagnosticSeverityGQL = DiagnosticSeverityGQL.Error,
        raise_error: Type[Exception] | None = None,
        uri: str | None = None,
    ):
        if not LSPErrorBuilder.lsp:
            # Don't do anything special let's just add the diagnostic
            return self.add_diagnostic(
                message=message,
                label=label,
                code=code,
                range=range,
                code_href=code_href,
                severity=severity,
                raise_error=raise_error,
                uri=uri,
            )

        candidates = [f"'{c}'" for c in spellcheck_candidates if not c.split(".")[-1].startswith("_")]
        if len(candidates) > 0:
            all_scores = [
                (
                    difflib.SequenceMatcher(a=spellcheck_item, b=candidate).quick_ratio(),
                    candidate,
                )
                for candidate in candidates
            ]
            all_scores.sort(key=lambda x: -x[0])

            if len(candidates) > 5:
                prefix = "The closest options are"
                candidates = [c for (_, c) in all_scores[:5]]
            elif len(candidates) == 1:
                prefix = "The only valid option is"
            else:
                prefix = "Valid options are"
                candidates = [c for (_, c) in all_scores]

            message += f" {prefix} {oxford_comma_list(candidates)}."

        return self.add_diagnostic(
            message=message,
            label=label,
            code=code,
            range=range,
            code_href=code_href,
            severity=severity,
            raise_error=raise_error,
            uri=uri,
        )


"""
The following helper methods are empirically proven to be correct.
The lines are 0 indexed and inclusive.
The character offsets are 1 indexed and non-inclusive
"""


def range_or_node_to_start_line(range_or_node: RangeGQL | ast.AST) -> int | None:
    if isinstance(range_or_node, RangeGQL):
        i = range_or_node.start.line
    else:
        i = getattr(range_or_node, "lineno", None)
        if i is None:
            return None
    return i - 1


def range_or_node_to_end_line(range_or_node: RangeGQL | ast.AST) -> int | None:
    if isinstance(range_or_node, RangeGQL):
        i = range_or_node.end.line
    else:
        i = getattr(range_or_node, "end_lineno", None)
        if i is None:
            return None
    return i - 1


def range_or_node_to_start_char(range_or_node: RangeGQL | ast.AST) -> int | None:
    if isinstance(range_or_node, RangeGQL):
        i = range_or_node.start.character
    else:
        i = getattr(range_or_node, "col_offset", None)
        if i is None:
            return None
    return i


def range_or_node_to_end_char(range_or_node: RangeGQL | ast.AST) -> int | None:
    if isinstance(range_or_node, RangeGQL):
        i = range_or_node.end.character
    else:
        i = getattr(range_or_node, "end_col_offset", None)
        if i is None:
            return None
    return max(i - 1, 0)



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor

class OperationKind(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    OPERATION_KIND_UNSPECIFIED: _ClassVar[OperationKind]
    OPERATION_KIND_ONLINE_QUERY: _ClassVar[OperationKind]
    OPERATION_KIND_MIGRATION: _ClassVar[OperationKind]
    OPERATION_KIND_CRON: _ClassVar[OperationKind]
    OPERATION_KIND_STREAMING: _ClassVar[OperationKind]
    OPERATION_KIND_MIGRATION_SAMPLER: _ClassVar[OperationKind]
    OPERATION_KIND_WINDOWED_STREAMING: _ClassVar[OperationKind]
    OPERATION_KIND_OFFLINE_QUERY: _ClassVar[OperationKind]

OPERATION_KIND_UNSPECIFIED: OperationKind
OPERATION_KIND_ONLINE_QUERY: OperationKind
OPERATION_KIND_MIGRATION: OperationKind
OPERATION_KIND_CRON: OperationKind
OPERATION_KIND_STREAMING: OperationKind
OPERATION_KIND_MIGRATION_SAMPLER: OperationKind
OPERATION_KIND_WINDOWED_STREAMING: OperationKind
OPERATION_KIND_OFFLINE_QUERY: OperationKind



from __future__ import annotations

import dataclasses
import warnings
from collections import defaultdict
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

import pyarrow as pa

from chalk.features import DataFrame, Feature, Features
from chalk.sql._internal.incremental import IncrementalSettings
from chalk.sql._internal.query_execution_parameters import (
    QueryExecutionParameters,
    query_execution_parameters_from_env_vars,
)
from chalk.sql.protocols import BaseSQLSourceProtocol
from chalk.utils.async_helpers import async_enumerate
from chalk.utils.df_utils import pa_table_to_pl_df
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.string import normalize_string_for_matching

if TYPE_CHECKING:
    import sqlalchemy.ext.asyncio
    from sqlalchemy import Table
    from sqlalchemy.engine import Connection
    from sqlalchemy.sql import Select
    from sqlalchemy.sql.ddl import CreateTable, DropTable
    from sqlalchemy.sql.elements import TextClause


class Finalizer(str, Enum):
    ONE_OR_NONE = "OneOrNone"
    ONE = "One"
    FIRST = "First"
    ALL = "All"


def _get_matching_root_fqn(
    normalized_to_original_col: Mapping[str, str], expected_feature_root_fqns: Collection[str]
) -> dict[str, str]:
    # Returns a map from original_col to the matched columns
    candidates: List[str] = []
    lookup: defaultdict[str, list[str]] = defaultdict(list)
    for x in expected_feature_root_fqns:
        root_fqn_normalized = normalize_string_for_matching(x.lower())
        without_root_ns = normalize_string_for_matching(".".join(x.split(".")[1:]))
        lookup[root_fqn_normalized].append(x)
        lookup[without_root_ns].append(x)
        if len(x.split("@")) == 2:
            non_versioned = x.split("@")[0]
            non_versioned_normalized = normalize_string_for_matching(non_versioned.lower())
            non_versioned_without_root_ns = normalize_string_for_matching(".".join(non_versioned.split(".")[1:]))
            lookup[non_versioned_normalized].append(x)
            lookup[non_versioned_without_root_ns].append(x)

    new_mapped_cols = {}
    for normalized_col_name, original_col_name in normalized_to_original_col.items():
        if len(original_col_name.split("@")) == 2:
            version_number = original_col_name.split("@")[1]
            normalized_col_name = normalized_col_name[: -len(f"{version_number}")]
        candidates = lookup[normalized_col_name]
        if len(candidates) == 0:
            continue
        if len(candidates) > 1:
            # We really shouldn't hit this, unless if features are case-sensitive and the user is querying
            # snowflake which is case-insensitive
            raise ValueError(
                (
                    f"Column '{normalized_col_name}' was ambiguous which feature it referred to. "
                    f"Possible candidates were: {candidates}"
                )
            )
        new_mapped_cols[original_col_name] = candidates[0]
    return new_mapped_cols


class FinalizedChalkQuery:
    """A query that cannot be further filtered."""

    def __init__(
        self,
        query: Union[Select, TextClause],
        params: Mapping[str, Any],
        finalizer: Finalizer,
        incremental_settings: Optional[IncrementalSettings],
        source: BaseSQLSourceProtocol,
        fields: Mapping[str, Feature],
        temp_tables: (
            Mapping[str, Tuple[Mapping[str, sqlalchemy.types.TypeEngine], pa.Table, CreateTable, Table, DropTable]]
            | None
        ) = None,
        is_empty: bool = False,
    ) -> None:
        """
        :param fields: may be set to None, in which case it can be inferred from the output type of the resolver which returned this query using infer_fields_from_resolver_output
        """
        super().__init__()
        self._query = query
        self._params = dict(params)
        self._finalizer = finalizer
        self._incremental_settings = incremental_settings
        self._source = source
        self._fields = fields
        self._temp_tables = dict(temp_tables) if temp_tables is not None else {}
        self._is_empty = is_empty

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        raise RuntimeError(
            "FinalizedChalkQuery should not be subclassed. The Chalk runtime assumes that this class is final."
        )

    @property
    def incremental_settings(self):
        return self._incremental_settings

    @property
    def finalizer(self):
        return self._finalizer

    @property
    def query(self):
        return self._query

    @property
    def source(self):
        return self._source

    @property
    def fields(self):
        return self._fields

    @property
    def params(self):
        return self._params

    @property
    def temp_tables(self):
        return self._temp_tables

    @property
    def is_empty(self):
        return self._is_empty

    def infer_fields_from_resolver_outputs(self, expected_features: Sequence[Feature[Any, Any]] | None):
        if len(self._fields) > 0:
            return self
        if expected_features is None:
            return self
        new_fields = {f.name: f for f in expected_features}
        return FinalizedChalkQuery(
            query=self._query,
            params=self._params,
            finalizer=self._finalizer,
            incremental_settings=self._incremental_settings,
            source=self._source,
            fields=new_fields,
            temp_tables=self._temp_tables,
            is_empty=self._is_empty,
        )

    def execute(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[Any] = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> Union[Features, DataFrame, None]:
        """Actually execute the query to a `DataFrame` or set of features.

        If the finalizer was ONE, ONE_OR_NONE, or FIRST, then a `Features` instance
        is returned. Otherwise, if the `finalizer` is ALL, then a `DataFrame` instance
        is returned.

        Parameters
        ----------
        expected_features
            The list of expected features for the output, as provided by the resovler in which this query is executed.
            If not specified, the column names as returned by the query will be used to determine the output features.
        connection
            Execute the query using the supplied connection. If `None` (default), a new connection will be acquired
            from the underlying source for the query
        query_execution_parameters
            Parameters to decide method of efficient/inefficient execution

        Returns
        -------
        DataFrame
            A `DataFrame`, if the `.finalizer` is ALL; otherwise, `Features` set.
            If the finalizer is ONE_OR_NONE or FIRST, then the result may be
            `None` if no row was found
        """
        return self._pa_table_to_res(
            self.execute_to_pyarrow(expected_features, connection, query_execution_parameters),
            expected_features,
        )

    async def async_execute(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[Any] = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> Union[Features, DataFrame, None]:
        """Actually execute the query to a `DataFrame` or set of features.

        If the finalizer was ONE, ONE_OR_NONE, or FIRST, then a `Features` instance
        is returned. Otherwise, if the `finalizer` is ALL, then a `DataFrame` instance
        is returned.

        Parameters
        ----------
        expected_features
            The list of expected features for the output, as provided by the resovler in which this query is executed.
            If not specified, the column names as returned by the query will be used to determine the output features.
        connection
            Execute the query using the supplied connection. If `None` (default), a new connection will be acquired
            from the underlying source for the query
        query_execution_parameters
            Parameters to decide method of efficient/inefficient execution

        Returns
        -------
        DataFrame
            A `DataFrame`, if the `.finalizer` is ALL; otherwise, `Features` set.
            If the finalizer is ONE_OR_NONE or FIRST, then the result may be
            `None` if no row was found
        """
        return self._pa_table_to_res(
            await self.async_execute_to_pyarrow(expected_features, connection, query_execution_parameters),
            expected_features,
        )

    def _pa_table_to_res(self, pa_table: pa.RecordBatch | pa.Table, expected_features: Optional[Sequence[Feature]]):
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")
        pa_table = self._postprocess_table(0, pa_table)
        if isinstance(pa_table, pa.RecordBatch):
            pa_table = pa.Table.from_batches([pa_table])
        df = pa_table_to_pl_df(pa_table)
        assert isinstance(df, pl.DataFrame)
        if self._finalizer in (Finalizer.ONE_OR_NONE, Finalizer.FIRST) and len(df) == 0:
            return None
        res = DataFrame(df)
        if self._finalizer in (Finalizer.ONE, Finalizer.ONE_OR_NONE, Finalizer.FIRST):
            return res.slice(0, 1).to_features()[0]
        return res

    def _get_col_to_feature(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
    ):
        expected_feature_root_fqns = (
            None if expected_features is None else frozenset(x.root_fqn for x in expected_features)
        )

        def col_to_features(column_names: Sequence[str]) -> Mapping[str, Feature]:
            # First map the column names to determine the feature fqns
            col_name_mapping = self.get_col_name_mapping(tuple(column_names), expected_feature_root_fqns)
            return {k: Feature.from_root_fqn(v) for (k, v) in col_name_mapping.items()}

        return col_to_features

    def _check_incremental_settings(self):
        if self.incremental_settings is not None:
            # FIXME: Move the incrementalization logic here, so then the `execute` and `execute_to_dataframe`
            # methods can take the hwm timestamp as a parameter, to allow for direct execution
            warnings.warn(
                (
                    "This query specified an incremental configuration, which has not been applied. "
                    "This is likely because the resolver is being executed directly. "
                    "The query will be attempted without any high-water-mark timestamp. "
                    "This will attempt to select all data, "
                    "or if the filters depend on the incremental timestamp, "
                    "will result in a query execution error. "
                )
            )

    def execute_to_pyarrow(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[Connection] = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> pa.Table:
        """Actually execute the query, and return a PyArrow table. Unlike :meth:`.execute`, this method will always keep the
        results as a PyArrow table, even if the finalizer implies a singleton results (e.g. ONE, ONE_OR_NONE, or FIRST).

        Parameters
        ----------
        expected_features
            The list of expected features for the output, as provided by the resolver in which this query is executed.
            If not specified, the column names as returned by the query will be used to determine the output features.
        connection
            Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
            from the underlying source for the query
        query_execution_parameters
            Optional, but should be set if called from the engine. Contains options that decide the query execution
            codepath.

        Returns
        -------
        Table
            A `pa.Table`, even if the result contains 0 or 1 rows.
        """
        if query_execution_parameters is None:
            query_execution_parameters = query_execution_parameters_from_env_vars()
        query_execution_parameters = dataclasses.replace(query_execution_parameters, yield_empty_batches=True)
        batches = list(self.execute_to_pyarrow_batches(expected_features, connection, query_execution_parameters))
        if len(batches) == 0:
            # This should be impossible to reach because we specify yield_empty_batches=True, so batches should generally contain at least one element
            if expected_features is None:
                # We have no idea what the expected schema is. This should
                return pa.Table.from_pydict({})
            else:
                return pa.Table.from_pydict(
                    {k.root_fqn: [] for k in expected_features},
                    pa.schema({k.root_fqn: k.converter.pyarrow_dtype for k in expected_features}),
                )
        return pa.Table.from_batches(batches)

    def execute_to_pyarrow_batches(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[Connection] = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> Iterable[pa.RecordBatch]:
        """Actually execute the query, and return a stream of `pa.RecordBatch`.

        Parameters
        ----------
        expected_features
            The list of expected features for the output, as provided by the resolver in which this query is executed.
            If not specified, the column names as returned by the query will be used to determine the output features.
        connection
            Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
            from the underlying source for the query
        query_execution_parameters
            Optional, but should be set if called from the engine. Contains options that decide the query execution
            codepath.

        Yields
        ------
        A stream of `pa.RecordBatch`
        """
        from chalk.sql._internal.sql_source import BaseSQLSource
        from chalk.sql._internal.sql_source_group import SQLSourceGroup

        self._check_incremental_settings()
        col_to_features = self._get_col_to_feature(expected_features)
        assert isinstance(self.source, (BaseSQLSource, SQLSourceGroup)), f"Expected BaseSQLSource, got {self.source}"

        if query_execution_parameters is None:
            query_execution_parameters = query_execution_parameters_from_env_vars()

        total_row_count = 0

        for i, pa_table in enumerate(
            self.source.execute_query(
                finalized_query=self,
                columns_to_features=col_to_features,
                connection=connection,
                query_execution_parameters=query_execution_parameters,
            )
        ):
            assert (
                len(pa_table) > 0 or query_execution_parameters.yield_empty_batches
            ), "Should never yield empty batches"
            pa_table = self._postprocess_table(i, pa_table)
            total_row_count += len(pa_table)
            yield pa_table
            if self.finalizer == Finalizer.FIRST and total_row_count > 0:
                break
        if total_row_count == 0 and self.finalizer == Finalizer.ONE:
            raise ValueError("Expected exactly one row; got 0 rows")

    async def async_execute_to_pyarrow(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection] = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> pa.Table:
        """Actually execute the query, and return a PyArrow table. Unlike :meth:`.execute`, this method will always keep the
        results as a PyArrow table, even if the finalizer implies a singleton results (e.g. ONE, ONE_OR_NONE, or FIRST).

        Parameters
        ----------
        expected_features
            The list of expected features for the output, as provided by the resolver in which this query is executed.
            If not specified, the column names as returned by the query will be used to determine the output features.
        connection
            Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
            from the underlying source for the query
        query_execution_parameters
            Optional, but should be set if called from the engine. Contains options that decide the query execution
            codepath.

        Returns
        -------
        Table
            A `pa.Table`, even if the result contains 0 or 1 rows.
        """
        if query_execution_parameters is None:
            query_execution_parameters = query_execution_parameters_from_env_vars()
        query_execution_parameters = dataclasses.replace(query_execution_parameters, yield_empty_batches=True)
        batches: list[pa.RecordBatch] = []
        async for batch in self.async_execute_to_pyarrow_batches(
            expected_features, connection, query_execution_parameters
        ):
            batches.append(batch)
        if len(batches) == 0:
            # This should be impossible to reach because we specify yield_empty_batches=True, so batches should generally contain at least one element
            if expected_features is None:
                # We have no idea what the expected schema is
                return pa.Table.from_pydict({})
            else:
                return pa.Table.from_pydict(
                    {k.root_fqn: [] for k in expected_features},
                    pa.schema({k.root_fqn: k.converter.pyarrow_dtype for k in expected_features}),
                )
        return pa.Table.from_batches(batches)

    async def async_execute_to_pyarrow_batches(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection] = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ) -> AsyncIterable[pa.RecordBatch]:
        """Actually execute the query, and return a stream of pa.RecordBatches.

        Parameters
        ----------
        expected_features
            The list of expected features for the output, as provided by the resolver in which this query is executed.
            If not specified, the column names as returned by the query will be used to determine the output features.
        connection
            Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
            from the underlying source for the query
        query_execution_parameters
            Optional, but should be set if called from the engine. Contains options that decide the query execution
            codepath.

        Yields
        ------
        A stream of `pa.RecordBatch`
        """
        from chalk.sql._internal.sql_source import BaseSQLSource

        if query_execution_parameters is None:
            query_execution_parameters = query_execution_parameters_from_env_vars()

        self._check_incremental_settings()
        col_to_features = self._get_col_to_feature(expected_features)
        assert isinstance(self.source, BaseSQLSource)

        total_row_count = 0

        async for i, pa_table in async_enumerate(
            self.source.async_execute_query(self, col_to_features, connection, query_execution_parameters)
        ):
            assert (
                len(pa_table) > 0 or query_execution_parameters.yield_empty_batches
            ), "Should never yield empty batches"
            pa_table = self._postprocess_table(i, pa_table)
            total_row_count += len(pa_table)
            yield pa_table
            if self.finalizer == Finalizer.FIRST and total_row_count > 0:
                break

        if total_row_count == 0 and self.finalizer == Finalizer.ONE:
            raise ValueError("Expected exactly one row; got 0 rows")

    @overload
    def _postprocess_table(self, i: int, pa_table: pa.RecordBatch) -> pa.RecordBatch:
        ...

    @overload
    def _postprocess_table(self, i: int, pa_table: pa.Table) -> pa.Table:  # pyright: ignore[reportOverlappingOverload]
        ...

    def _postprocess_table(self, i: int, pa_table: pa.RecordBatch | pa.Table) -> pa.RecordBatch | pa.Table:
        if len(pa_table) == 0:
            return pa_table
        if i > 0:
            if self.finalizer == Finalizer.ONE_OR_NONE:
                raise ValueError(f"Expected zero or one rows, but got multiple batches")
            if self.finalizer == Finalizer.ONE:
                raise ValueError(f"Expected one row, but got multiple batches")
            assert self.finalizer != Finalizer.FIRST
        if self._finalizer == Finalizer.ONE:
            if len(pa_table) != 1:
                raise ValueError(f"Expected exactly one row; got {len(pa_table)} rows")

        if self._finalizer == Finalizer.ONE_OR_NONE:
            if len(pa_table) > 1:
                raise ValueError(f"Expected zero or one rows; got {len(pa_table)} rows")

        if self._finalizer in (Finalizer.ONE, Finalizer.ONE_OR_NONE, Finalizer.FIRST):
            pa_table = pa_table.slice(0, 1)
        return pa_table

    def execute_to_dataframe(
        self,
        expected_features: Optional[Sequence[Feature]] = None,
        connection: Optional[Connection] = None,
        query_execution_parameters: Optional[QueryExecutionParameters] = None,
    ):
        # DEPRECATED -- will be removed soon. Only used internally
        """Actually execute the query, and return a DataFrame. Unlike :meth:`.execute`, this method will always keep the
        results as a DataFrame, even if the finalizer implies a singleton results (e.g. ONE, ONE_OR_NONE, or FIRST).

        Parameters
        ----------
        expected_features
            The list of expected features for the output, as provided by the resovler in which this query is executed.
            If not specified, the column names as returned by the query will be used to determine the output features.
        connection
            Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
            from the underlying source for the query
        query_execution_parameters
            Parameters to decide method of efficient/inefficient execution

        Returns
        -------
        DataFrame
            A `DataFrame`, even if the result contains 0 or 1 rows.
        """
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")
        pa_table = self.execute_to_pyarrow(expected_features, connection, query_execution_parameters)
        df = pa_table_to_pl_df(pa_table)
        assert isinstance(df, pl.DataFrame)
        return DataFrame(df)

    def get_col_name_mapping(
        self,
        result_columns: Tuple[str, ...],
        expected_features: Optional[Collection[str]],
    ) -> Dict[str, str]:
        """Map the output columns to the expected feature names.

        Parameters
        ----------
        result_columns
            A list of the columns, in order, returned by the query.
        expected_features
            The expected feature root fqns for the query, as provided by the resolver signature.
            If a column name that is not in `fields` corresponds to a column in `expected_features`,
            then it will be mapped automatically.
            If a feature in `expected_features` does not have a corresponding output column,
            an error is raised.

        Returns
        -------
        dict[str, str]
            A mapping from output column names to root names.
        """
        ans: Dict[str, str] = {}
        normalized_to_original_col = {normalize_string_for_matching(x): x for x in result_columns}
        for k, v in self._fields.items():
            original_col_name = normalized_to_original_col.get(normalize_string_for_matching(k))
            if original_col_name is None:
                received_columns = ", ".join(result_columns)
                raise ValueError(
                    f"Required column '{k}' was not returned by the query. Got columns: {received_columns}"
                )
            ans[original_col_name] = v.root_fqn
        if expected_features is not None:
            unexpected_fields = [x for x in self._fields.values() if x.root_fqn not in expected_features]
            for x in unexpected_fields:
                del ans[normalized_to_original_col[normalize_string_for_matching(x.root_fqn)]]
            unsolved_normalized_to_original_col = {
                normalized_col: original_col
                for normalized_col, original_col in normalized_to_original_col.items()
                if original_col not in ans
            }
            ans.update(_get_matching_root_fqn(unsolved_normalized_to_original_col, expected_features))
        return ans


if TYPE_CHECKING:

    class SingletonFinalizedChalkQuery(FinalizedChalkQuery, Features):
        """A FinalizedChalkQuery that returns a single row when executed"""

        # Subclassing from Features so it can does not cause type errors when used in a resolver
        # that is annotated to return a Features instance
        ...

    class DataframeFinalizedChalkQuery(FinalizedChalkQuery, DataFrame):
        """A FinalizedChalkQuery that returns a DataFrame when executed"""

        # Subclassing from DataFrame so it can does not cause type errors when used in a resolver
        # that is annotated to return a DataFrame
        ...

else:
    SingletonFinalizedChalkQuery = FinalizedChalkQuery
    DataframeFinalizedChalkQuery = FinalizedChalkQuery



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor

class QueryStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    QUERY_STATUS_UNSPECIFIED: _ClassVar[QueryStatus]
    QUERY_STATUS_PENDING_SUBMISSION: _ClassVar[QueryStatus]
    QUERY_STATUS_SUBMITTED: _ClassVar[QueryStatus]
    QUERY_STATUS_RUNNING: _ClassVar[QueryStatus]
    QUERY_STATUS_ERROR: _ClassVar[QueryStatus]
    QUERY_STATUS_EXPIRED: _ClassVar[QueryStatus]
    QUERY_STATUS_CANCELLED: _ClassVar[QueryStatus]
    QUERY_STATUS_SUCCESSFUL: _ClassVar[QueryStatus]
    QUERY_STATUS_SUCCESSFUL_WITH_NONFATAL_ERRORS: _ClassVar[QueryStatus]

QUERY_STATUS_UNSPECIFIED: QueryStatus
QUERY_STATUS_PENDING_SUBMISSION: QueryStatus
QUERY_STATUS_SUBMITTED: QueryStatus
QUERY_STATUS_RUNNING: QueryStatus
QUERY_STATUS_ERROR: QueryStatus
QUERY_STATUS_EXPIRED: QueryStatus
QUERY_STATUS_CANCELLED: QueryStatus
QUERY_STATUS_SUCCESSFUL: QueryStatus
QUERY_STATUS_SUCCESSFUL_WITH_NONFATAL_ERRORS: QueryStatus
